package de.hft.stuttgart.citydoctor2.gui;

import de.hft.stuttgart.citydoctor2.gui.tree.ZipEntryListCell;
import de.hft.stuttgart.citydoctor2.gui.tree.ZipEntryNode;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipArchive;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipEntry;
import de.hft.stuttgart.citydoctor2.zip.ZipEntryErrorType;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.ListView;
import javafx.scene.control.ProgressBar;
import javafx.scene.control.TextField;
import javafx.scene.control.TitledPane;
import javafx.scene.control.Tooltip;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.VBox;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.stage.Window;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.IOException;
import java.io.InputStream;

public class ZipEntryPicker {

    private static final Logger logger = LogManager.getLogger(ZipEntryPicker.class);

    private Stage stage;
    private CityGmlZipArchive archive;

    @FXML
    private TitledPane entriesPane;
    @FXML
    private TitledPane metadataPane;

    @FXML
    private Label subpathLbl;
    @FXML
    private TextField subpathValueTxt;

    @FXML
    private Label erroneousLbl;
    @FXML
    private Label erroneousValue;

    @FXML
    private Label filesizeLbl;
    @FXML
    private Label filesizeValue;

    @FXML
    private Label validatedLbl;
    @FXML
    private Label validatedValue;

    @FXML
    private Label objectCountLbl;
    @FXML
    private Label objectCountValue;

    @FXML
    private Button cancelBtn;

    @FXML
    private Button loadBtn;
    @FXML
    private ImageView loadImageView;

    @FXML
    private Button decompressBtn;
    @FXML
    private ImageView decompressImageView;

    @FXML
    private Button decompressAllBtn;
    @FXML
    private ImageView decompressAllImageView;

    @FXML
    private ListView<ZipEntryNode> entryList;
    @FXML
    private ProgressBar progress;
    @FXML
    private ProgressBar subProgress;

    private CityDoctorController controller;


    private String unknownValueText = Localization.getText("ZipEntryPicker.unknownValue");

    public ZipEntryPicker(Window parent, CityDoctorController controller) throws IOException {
        FXMLLoader loader = new FXMLLoader(ZipEntryPicker.class.getResource("ZipEntryPicker.fxml"));
        loader.setController(this);
        this.controller = controller;
        archive = controller.getZipArchive();
        VBox box = loader.load();
        stage = new Stage();
        stage.getIcons().add(new Image(MainWindow.class.getResourceAsStream("icons/CityDoctor-Logo-rot_klein.jpg")));
        stage.setScene(new Scene(box));
        stage.initOwner(parent);
        stage.initModality(Modality.APPLICATION_MODAL);
        stage.setTitle("Zip-Entry Picker");
        stage.getScene().addEventFilter(KeyEvent.KEY_PRESSED, (KeyEvent event) -> {
            if (event.getCode() == KeyCode.ESCAPE) {
                stage.close();
            }
        });
    }

    public void initialize() {
        subProgress.setVisible(false);
        entryList.getSelectionModel().selectedItemProperty().addListener((obs, oldI, newI) -> {
            if (newI != null) {
                showMetadata(newI.getEntry());
            }
        });
        entryList.setCellFactory(param -> new ZipEntryListCell());
        setupButtons();
        applyLocalization();
        populateZipEntryList();
    }

    private void setupButtons() {
        try {
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/decompress_zipEntry.png")) {
                Image img = new Image(inStream);
                decompressImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/load_zipEntry.png")) {
                Image img = new Image(inStream);
                loadImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/decompress_all_zipEntries.png")) {
                Image img = new Image(inStream);
                decompressAllImageView.setImage(img);
            }
        } catch (IOException ignored) {
            //ignore exceptions
        }

        decompressBtn.setOnAction(e -> {
            disableTaskButtons();
            Thread t = new Thread(() -> {
                CityGmlZipEntry entry = entryList.getSelectionModel().getSelectedItem().getEntry();
                controller.decompressZipEntry(entry, progressValue ->
                        Platform.runLater(() -> progress.setProgress(progressValue))
                );
                Platform.runLater(() -> {
                    if (!decompressAllBtn.isDisable() &&
                            entryList.getItems().stream().allMatch(zipEntryNode -> zipEntryNode.getEntry().isDecompressed())) {
                        decompressAllBtn.setDisable(true);
                    }
                    populateZipEntryList(entryList.getSelectionModel().getSelectedIndex());
                });
            });
            t.start();

        });

        loadBtn.setOnAction(e -> {
            disableTaskButtons();
            CityGmlZipEntry entry = entryList.getSelectionModel().getSelectedItem().getEntry();

        });

        decompressAllBtn.setOnAction(e -> {
            disableTaskButtons();
            subProgress.setVisible(true);
            subProgress.setProgress(0);
            Thread t = new Thread(() -> {
                progress.setProgress(0);
                int count = 0;
                int total = entryList.getItems().size();
                for (ZipEntryNode node : entryList.getItems()) {
                    controller.decompressZipEntry(node.getEntry(), progressValue ->
                            Platform.runLater(() -> subProgress.setProgress(progressValue))
                    );
                    count++;
                    double progressValue = count * 1.0 / total;
                    Platform.runLater(() -> progress.setProgress(progressValue));
                }
                Platform.runLater(() -> {
                    subProgress.setVisible(false);
                    populateZipEntryList(entryList.getSelectionModel().getSelectedIndex());
                });
            });
            t.start();
        });

        cancelBtn.setOnAction(e -> stage.close());
    }

    private void applyLocalization() {
        entriesPane.setText(Localization.getText("ZipEntryPicker.entryListTitle"));
        metadataPane.setText(Localization.getText("ZipEntryPicker.metadata"));
        subpathLbl.setText(Localization.getText("ZipEntryPicker.subpathLbl"));
        erroneousLbl.setText(Localization.getText("ZipEntryPicker.erroneousLbl"));
        filesizeLbl.setText(Localization.getText("ZipEntryPicker.filesizeLbl"));
        validatedLbl.setText(Localization.getText("ZipEntryPicker.validatedLbl"));
        objectCountLbl.setText(Localization.getText("ZipEntryPicker.objectCountLbl"));

        subpathValueTxt.setText(unknownValueText);
        erroneousValue.setText(unknownValueText);
        filesizeValue.setText(unknownValueText);
        validatedValue.setText(unknownValueText);
        objectCountValue.setText(unknownValueText);

        loadBtn.setTooltip(new Tooltip(Localization.getText("ZipEntryPicker.loadBtn")));
        decompressBtn.setTooltip(new Tooltip(Localization.getText("ZipEntryPicker.decompressBtn")));
        cancelBtn.setText(Localization.getText("ZipEntryPicker.cancelBtn"));
    }

    private void populateZipEntryList() {
        populateZipEntryList(0);
    }

    private void populateZipEntryList(int selectionIndex) {
        entryList.getItems().clear();
        for (CityGmlZipEntry entry : archive.getEntries()) {
            entryList.getItems().add(new ZipEntryNode(entry));
        }
        entryList.getSelectionModel().clearSelection();
        entryList.getSelectionModel().select(selectionIndex);
        showMetadata(entryList.getSelectionModel().getSelectedItem().getEntry());
    }

    private void showMetadata(CityGmlZipEntry entry) {
        subpathValueTxt.setText(entry.getFileName());
        if (entry.getFileSize() != -1L) {
            long fileSize = entry.getFileSize();
            long kb = 1024L;
            long mb = kb * kb;
            long gb = mb * kb;

            if (fileSize < kb) {
                filesizeValue.setText("< 1 KB");
            } else if (fileSize < mb) {
                filesizeValue.setText(String.format("%s KB", fileSize / kb));
            } else if (fileSize < gb) {
                filesizeValue.setText(String.format("%s MB", fileSize / mb));
            } else {
                double gigabytes = (double) fileSize / gb;
                filesizeValue.setText(String.format("%.2f GB", gigabytes));
            }
        } else {
            filesizeValue.setText(unknownValueText);
        }
        if (entry.getErrorType() != null) {
            erroneousValue.setText(getErrorText(entry.getErrorType()));
            validatedValue.setText(unknownValueText);
            objectCountValue.setText(unknownValueText);
            loadBtn.setDisable(true);
            decompressBtn.setDisable(true);
        } else {
            erroneousValue.setText(Localization.getText("ZipEntryPicker.no"));

            loadBtn.setDisable(!entry.isDecompressed());
            decompressBtn.setDisable(entry.isDecompressed());
            if (entry.getModel() == null) {
                validatedValue.setText(unknownValueText);
                objectCountValue.setText(unknownValueText);
                return;
            }
            if (entry.getModel().isValidated()) {
                validatedValue.setText(Localization.getText("ZipEntryPicker.yes"));
            } else {
                validatedValue.setText(Localization.getText("ZipEntryPicker.no"));
            }
            objectCountValue.setText(String.valueOf(entry.getModel().getNumberOfFeatures()));
        }
    }

    private void disableTaskButtons() {
        decompressBtn.setDisable(true);
        loadBtn.setDisable(true);
        decompressAllBtn.setDisable(true);
    }

    private String getErrorText(ZipEntryErrorType error) {
        switch (error) {
            case EXCESSIVE_FILESIZE -> {
                return Localization.getText("ZipEntryPicker.excessiveFileSize");
            }
            case INVALID_CITY_GML_FILE -> {
                return Localization.getText("ZipEntryPicker.invalidCityGml");
            }
            default -> {
                return Localization.getText("ZipEntryPicker.ioError");
            }
        }
    }

    public void show() {
        Platform.runLater(() -> {
            stage.showAndWait();
        });
    }
}
